"""
Tests for aspif parsing.
"""
from unittest import TestCase
from textwrap import dedent

from ..control import Control
from .util import solve


def solve_aspif(prg):
    ctl = Control()
    ctl.add(dedent(prg))
    return solve(ctl)


def theory(prg):
    ctl = Control()
    ctl.add(dedent(prg))
    tab = {}
    for atm in ctl.symbolic_atoms:
        tab[atm.literal] = str(atm.symbol)

    thy = []
    for x in ctl.theory_atoms:
        elems = []
        for elem in x.elements:
            terms = ", ".join(str(term) for term in elem.terms)
            cond = ", ".join(sorted(tab[lit] for lit in elem.condition))
            elems.append(f"{terms} : {cond}")
        thy.append(f"&{x.term} {{ {'; '.join(sorted(elems))} }}")
        if x.guard is not None:
            thy[-1] += f" {x.guard[0]} {x.guard[1]}"
    return sorted(thy)


class TestASPIF(TestCase):
    """
    Test ASPIF parsing.
    """

    def test_preamble(self):
        """
        Test preamble parsing.
        """
        ctl = Control()
        ctl.add(
            dedent(
                """\
            asp 1 0 0
            0
            """
            )
        )
        ctl = Control([], lambda msg, code: None)
        self.assertRaises(
            RuntimeError,
            ctl.add,
            dedent(
                """\
            asp 1 0 0 unknown
            0
            """
            ),
        )
        ctl = Control([], lambda msg, code: None)
        ctl.add(
            dedent(
                """\
            asp 1 0 0 incremental
            0
            """
            )
        )
        self.assertRaises(
            RuntimeError,
            ctl.add,
            dedent(
                """\
            asp 1 0 0 incremental
            0
            """
            ),
        )
        ctl = Control([], lambda msg, code: None)
        self.assertRaises(
            RuntimeError,
            ctl.add,
            dedent(
                """\
            asp 1 0 0 incremental
            0
            0
            """
            ),
        )

    def test_rule(self):
        """
        Test rule parsing.
        """
        # a fact
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            1 0 1 1 0 0
            4 1 a 1 1
            0
            """
            ),
            [["a"]],
        )
        # a choice
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            1 1 1 1 0 0
            4 1 a 1 1
            0
            """
            ),
            [[], ["a"]],
        )
        # a rule with normal body
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            1 1 1 1 0 0
            1 0 1 2 0 1 1
            1 0 1 3 0 1 -2
            4 1 a 1 1
            4 1 b 1 2
            4 1 c 1 3
            0
            """
            ),
            [["a", "b"], ["c"]],
        )
        # a rule with weight body
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            1 1 2 1 2 0 0
            1 0 1 3 1 2 2 1 1 2 1
            1 0 0 0 1 -3
            4 1 a 1 1
            4 1 b 1 2
            0
            """
            ),
            [["a", "b"]],
        )

    def test_minimize(self):
        """
        Test minimize parsing.
        """
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            1 1 2 1 2 0 0
            1 0 0 0 1 -2
            1 0 0 0 1 -1
            2 1 1 1 1
            2 2 1 2 2
            4 1 a 1 1
            4 1 b 1 2
            0
            """
            ),
            [["a", "b"]],
        )

    def test_external(self):
        """
        Test external parsing.
        """
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            5 1 0
            4 1 a 1 1
            0
            """
            ),
            [[], ["a"]],
        )
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            5 1 1
            4 1 a 1 1
            0
            """
            ),
            [["a"]],
        )
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            5 1 2
            4 1 a 1 1
            0
            """
            ),
            [[]],
        )

        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            5 1 3
            4 1 a 1 1
            0
            """
            ),
            [[]],
        )

    def test_assume(self):
        """
        Test assumption parsing.
        """
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            1 1 2 1 2 0 0
            6 2 1 -2
            4 1 a 1 1
            4 1 b 1 2
            0
            """
            ),
            [["a"]],
        )

    def test_heuristic(self):
        """
        Test heuristic parsing.
        """
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            1 1 1 1 0 0
            7 4 1 1 2 0
            4 1 a 1 1
            0
            """
            ),
            [[], ["a"]],
        )

    def test_edge(self):
        """
        Test edge parsing.
        """
        self.assertEqual(
            solve_aspif(
                """\
            asp 1 0 0
            1 1 1 1 0 0
            8 0 1 1 1
            8 1 0 0
            4 1 a 1 1
            0
            """
            ),
            [[]],
        )

    def test_theory(self):
        """
        Test theory parsing.
        """
        # no guard
        self.assertEqual(
            theory(
                """\
            asp 1 0 0
            1 1 1 1 0 0
            1 0 1 2 0 0
            9 1 0 1 b
            9 0 1 1
            9 4 0 1 1 1 1
            9 5 2 0 1 0
            4 1 x 1 1
            0
            """
            ),
            ["&b { 1 : x }"],
        )
        # with guard
        self.assertEqual(
            theory(
                """\
            asp 1 0 0
            1 1 1 1 0 0
            1 0 1 2 0 0
            9 1 0 1 a
            9 0 3 1
            9 4 0 1 3 1 1
            9 1 2 3 <=<
            9 0 1 2
            9 6 2 0 1 0 2 1
            4 1 x 1 1
            0
            """
            ),
            ["&a { 1 : x } <=< 2"],
        )
        # complex terms
        self.assertEqual(
            theory(
                """\
            asp 1 0 0
            1 0 1 1 0 0
            9 1 0 1 b
            9 0 3 1
            9 0 4 2
            9 1 2 3 +-*
            9 2 5 2 2 3 4
            9 2 6 -1 2 3 4
            9 0 7 3
            9 0 8 4
            9 2 9 -3 2 7 8
            9 0 10 5
            9 0 11 6
            9 2 12 -2 2 10 11
            9 1 1 1 f
            9 2 13 1 4 5 6 9 12
            9 4 0 1 13 0
            9 5 1 0 1 0
            0
            """
            ),
            ["&b { f((1+-*2),(1,2),[3,4],{5,6}) :  }"],
        )

    def test_comment(self):
        """
        Test comment parsing.
        """
        ctl = Control()
        ctl.add(
            dedent(
                """\
            asp 1 0 0
            10 nothing!
            0
            """
            )
        )



"""
Module providing the clingo-like pyclingo application.
"""
from clingo import Application, clingo_main


class PyClingoApplication(Application):
    """
    This is an example app mimimicking clingo.
    """

    program_name = "pyclingo"

    def main(self, control, files):
        for file_ in files:
            control.load(file_)
        if not files:
            control.load("-")
        control.ground([("base", [])])
        control.solve()


if __name__ == "__main__":
    clingo_main(PyClingoApplication())



"""
Core functionality used throught the clingo package.
"""

from typing import Callable, Tuple
from enum import Enum

from ._internal import _cb_error_panic, _ffi, _lib, _to_str

__all__ = ["Logger", "MessageCode", "OrderedEnum", "TruthValue", "version"]


class OrderedEnum(Enum):
    """
    Enumeration of orderable elements.
    """

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


def version() -> Tuple[int, int, int]:
    """
    Clingo's version as a tuple `(major, minor, revision)`.
    """
    p_major = _ffi.new("int*")
    p_minor = _ffi.new("int*")
    p_revision = _ffi.new("int*")
    _lib.clingo_version(p_major, p_minor, p_revision)
    return p_major[0], p_minor[0], p_revision[0]


class MessageCode(OrderedEnum):
    """
    Enumeration of messages codes.
    """

    AtomUndefined = _lib.clingo_warning_atom_undefined
    """
    Informs about an undefined atom in program.
    """
    FileIncluded = _lib.clingo_warning_file_included
    """
    Indicates that the same file was included multiple times.
    """
    GlobalVariable = _lib.clingo_warning_global_variable
    """
    Informs about a global variable in a tuple of an aggregate element.
    """
    OperationUndefined = _lib.clingo_warning_operation_undefined
    """
    Inform about an undefined arithmetic operation or unsupported weight of an
    aggregate.
    """
    Other = _lib.clingo_warning_other
    """
    Reports other kinds of messages.
    """
    RuntimeError = _lib.clingo_warning_runtime_error
    """
    To report multiple errors; a corresponding runtime error is raised later.
    """
    VariableUnbounded = _lib.clingo_warning_variable_unbounded
    """
    Informs about a CSP variable with an unbounded domain.
    """


Logger = Callable[[MessageCode, str], None]


@_ffi.def_extern(onerror=_cb_error_panic, name="pyclingo_logger_callback")
def _pyclingo_logger_callback(code, message, data):
    """
    Low-level logger callback.
    """
    handler = _ffi.from_handle(data)
    handler(MessageCode(code), _to_str(message))


class TruthValue(OrderedEnum):
    """
    Enumeration of the different truth values.
    """

    False_ = _lib.clingo_external_type_false
    """
    Represents truth value true.
    """
    Free = _lib.clingo_external_type_free
    """
    Represents absence of a truth value.
    """
    True_ = _lib.clingo_external_type_true
    """
    Represents truth value true.
    """
    Release = _lib.clingo_external_type_release
    """
    Indicates that an atom is to be released.
    """



import sys
from argparse import ArgumentParser

if sys.version_info >= (3, 11):
    NEW_STYLE = True
    from sysconfig import get_config_var, get_config_vars, get_preferred_scheme, get_path
else:
    NEW_STYLE = False
    from site import USER_SITE
    from distutils.sysconfig import get_python_lib, get_config_vars

parser = ArgumentParser()
if sys.version_info >= (3, 7):
    subparser = parser.add_subparsers(required=True, dest="action")
else:
    subparser = parser.add_subparsers(dest="action")

prefix_parser = subparser.add_parser("target")
prefix_group = prefix_parser.add_mutually_exclusive_group()
prefix_group.add_argument("--user", action='store_true', help='get user prefix')
prefix_group.add_argument("--prefix", type=str, help='prepend prefix')

prefix_parser = subparser.add_parser("suffix")

result = parser.parse_args()

if NEW_STYLE:
    if result.action == "target":
        if result.user:
            scheme = get_preferred_scheme("user")
        else:
            scheme = get_preferred_scheme("prefix")
        if result.prefix is not None:
            cvars = get_config_vars()
            cvars["base"] = result.prefix
            cvars["platbase"] = result.prefix
            platlib = get_path("platlib", scheme, cvars)
        else:
            platlib = get_path("platlib", scheme)
        print(platlib)
    elif result.action == "suffix":
        print(get_config_var('EXT_SUFFIX'))
else:
    # TODO: remove once python 3.10 is eol
    if result.action == "target":
        if result.user:
            print(USER_SITE)
        else:
            print(get_python_lib(True, False, result.prefix))
    elif result.action == "suffix":
        SO, SOABI, EXT_SUFFIX = get_config_vars("SO", "SOABI", "EXT_SUFFIX")
        if EXT_SUFFIX is not None:
            ext = EXT_SUFFIX
        elif SOABI is not None:
            ext = ''.join('.', SOABI, SO)
        else:
            ext = SO
        print(ext)



"""
Helpers to simplify testing.
"""

from typing import List, cast
from unittest import TestCase

from ..control import Control
from ..configuration import Configuration
from .. import SolveResult, parse_term


def _p(*models):
    return [[parse_term(symbol) for symbol in model] for model in models]


class _MCB:
    # pylint: disable=missing-function-docstring
    def __init__(self):
        self._models = []
        self._core = None
        self.last = None

    def on_core(self, c):
        self._core = c

    def on_model(self, m):
        self.last = (m.type, sorted(m.symbols(shown=True)))
        self._models.append(self.last[1])

    @property
    def core(self):
        return sorted(self._core)

    @property
    def models(self):
        return sorted(self._models)


def _check_sat(case: TestCase, ret: SolveResult) -> None:
    case.assertTrue(ret.satisfiable is True)
    case.assertTrue(ret.unsatisfiable is False)
    case.assertTrue(ret.unknown is False)
    case.assertTrue(ret.exhausted is True)


def solve(ctl: Control) -> List[List[str]]:
    """
    Solve with the given control ojbect and return models as sorted lists of
    strings.
    """
    cast(Configuration, ctl.configuration.solve).models = "0"
    mcb = _MCB()
    ctl.solve(on_model=mcb.on_model)
    return [[str(s) for s in m] for m in mcb.models]



"""
Tests for the backend/observer.
"""

from unittest import TestCase
from typing import Sequence, Tuple
from clingo.symbol import parse_term
from clingo import (
    Control,
    Function,
    HeuristicType,
    Observer,
    Symbol,
    TheorySequenceType,
    TruthValue,
)


class TestObserverBackend(Observer):
    """
    Test Observer.
    """

    def __init__(self, case):
        self._case = case
        self.called = set()

    def init_program(self, incremental: bool) -> None:
        self.called.add("init_program")

    def begin_step(self) -> None:
        self.called.add("begin_step")

    def rule(self, choice: bool, head: Sequence[int], body: Sequence[int]) -> None:
        self.called.add("rule")
        self._case.assertTrue(choice)
        self._case.assertEqual(head, [1])
        self._case.assertEqual(body, [2, 3])

    def weight_rule(
        self,
        choice: bool,
        head: Sequence[int],
        lower_bound: int,
        body: Sequence[Tuple[int, int]],
    ) -> None:
        self.called.add("weight_rule")
        self._case.assertFalse(choice)
        self._case.assertEqual(head, [2])
        self._case.assertEqual(lower_bound, 1)
        self._case.assertEqual(body, [(2, 3), (4, 5)])

    def minimize(self, priority: int, literals: Sequence[Tuple[int, int]]) -> None:
        self.called.add("minimize")
        self._case.assertEqual(priority, 0)
        self._case.assertEqual(literals, [(2, 3), (4, 5)])

    def project(self, atoms: Sequence[int]) -> None:
        self.called.add("project")
        self._case.assertEqual(atoms, [2, 4])

    def output_atom(self, symbol: Symbol, atom: int) -> None:
        self.called.add("output_atom")
        self._case.assertEqual(symbol, Function("a"))
        self._case.assertEqual(atom, 2)

    def external(self, atom: int, value: TruthValue) -> None:
        self.called.add("external")
        self._case.assertEqual(atom, 3)
        self._case.assertEqual(value, TruthValue.Release)

    def assume(self, literals: Sequence[int]) -> None:
        self.called.add("assume")
        self._case.assertEqual(literals, [2, 3])

    def heuristic(
        self,
        atom: int,
        type_: HeuristicType,
        bias: int,
        priority: int,
        condition: Sequence[int],
    ) -> None:
        self.called.add("heuristic")
        self._case.assertEqual(atom, 2)
        self._case.assertEqual(type_, HeuristicType.Level)
        self._case.assertEqual(bias, 5)
        self._case.assertEqual(priority, 7)
        self._case.assertEqual(condition, [1, 3])

    def acyc_edge(self, node_u: int, node_v: int, condition: Sequence[int]) -> None:
        self.called.add("acyc_edge")
        self._case.assertEqual(node_u, 1)
        self._case.assertEqual(node_v, 2)
        self._case.assertEqual(condition, [3, 4])

    def end_step(self) -> None:
        self.called.add("end_step")


class TestObserverTheory(Observer):
    """
    Test Observer.
    """

    def __init__(self, case):
        self._case = case
        self.called = set()

    def output_term(self, symbol: Symbol, condition: Sequence[int]) -> None:
        self.called.add("output_term")
        self._case.assertEqual(symbol, Function("t"))
        self._case.assertGreaterEqual(len(condition), 1)

    def theory_term_number(self, term_id: int, number: int) -> None:
        self.called.add("theory_term_number")
        self._case.assertEqual(number, 1)

    def theory_term_string(self, term_id: int, name: str) -> None:
        self.called.add("theory_term_string")
        self._case.assertEqual(name, "a")

    def theory_term_compound(
        self, term_id: int, name_id_or_type: int, arguments: Sequence[int]
    ) -> None:
        self.called.add("theory_term_compound")
        self._case.assertEqual(name_id_or_type, -1)
        self._case.assertGreaterEqual(len(arguments), 2)

    def theory_element(
        self, element_id: int, terms: Sequence[int], condition: Sequence[int]
    ) -> None:
        self.called.add("theory_element")
        self._case.assertEqual(len(terms), 1)
        self._case.assertEqual(len(condition), 2)

    def theory_atom(
        self, atom_id_or_zero: int, term_id: int, elements: Sequence[int]
    ) -> None:
        self.called.add("theory_atom")
        self._case.assertEqual(len(elements), 1)


class TestObserverTheoryWithGuard(Observer):
    """
    Test Observer.
    """

    def __init__(self, case):
        self._case = case
        self.called = set()

    def theory_term_string(self, term_id: int, name: str) -> None:
        self.called.add(f"theory_term_string: {name}")

    def theory_atom_with_guard(
        self,
        atom_id_or_zero: int,
        term_id: int,
        elements: Sequence[int],
        operator_id: int,
        right_hand_side_id: int,
    ) -> None:
        self.called.add("theory_atom_with_guard")
        self._case.assertEqual(len(elements), 0)


class TestBackend(TestCase):
    """
    Tests basic solving and related functions.
    """

    def test_backend(self):
        """
        Test backend via observer.
        """
        ctl = Control()
        obs = TestObserverBackend(self)
        ctl.register_observer(obs)
        with ctl.backend() as backend:
            self.assertIn("init_program", obs.called)
            self.assertIn("begin_step", obs.called)
            backend.add_atom()
            backend.add_atom(Function("a"))
            backend.add_rule([1], [2, 3], True)
            self.assertIn("rule", obs.called)
            backend.add_weight_rule([2], 1, [(2, 3), (4, 5)])
            self.assertIn("weight_rule", obs.called)
            backend.add_minimize(0, [(2, 3), (4, 5)])
            self.assertIn("minimize", obs.called)
            backend.add_project([2, 4])
            self.assertIn("project", obs.called)
            backend.add_heuristic(2, HeuristicType.Level, 5, 7, [1, 3])
            self.assertIn("heuristic", obs.called)
            backend.add_assume([2, 3])
            self.assertIn("assume", obs.called)
            backend.add_acyc_edge(1, 2, [3, 4])
            self.assertIn("acyc_edge", obs.called)
            backend.add_external(3, TruthValue.Release)
            self.assertIn("external", obs.called)
        self.assertIn("output_atom", obs.called)
        ctl.solve()
        self.assertIn("end_step", obs.called)

    def test_theory(self):
        """
        Test observer via grounding.
        """
        ctl = Control()
        obs = TestObserverTheory(self)
        ctl.register_observer(obs)
        ctl.add(
            "base",
            [],
            """\
        #theory test {
            t { };
            &a/0 : t, head
        }.
        {a; b}.
        #show t : a, b.
        &a { (1,a): a,b }.
        """,
        )
        ctl.ground([("base", [])])
        self.assertIn("output_term", obs.called)
        self.assertIn("theory_term_number", obs.called)
        self.assertIn("theory_term_string", obs.called)
        self.assertIn("theory_term_compound", obs.called)
        self.assertIn("theory_element", obs.called)
        self.assertIn("theory_atom", obs.called)
        ctl.solve()

    def test_adding_theory(self):
        """
        Test theory related functions in backend.
        """
        ctl = Control()
        with ctl.backend() as backend:
            num_one = backend.add_theory_term_number(1)
            num_two = backend.add_theory_term_number(2)
            str_x = backend.add_theory_term_string("x")
            fun = backend.add_theory_term_function("f", [num_one, num_two, str_x])
            seq = backend.add_theory_term_sequence(
                TheorySequenceType.Set, [num_one, num_two, str_x]
            )
            lst = backend.add_theory_term_sequence(TheorySequenceType.List, [num_one, num_two])
            tup = backend.add_theory_term_sequence(TheorySequenceType.Tuple, [num_one, num_two])
            fseq = backend.add_theory_term_function("f", [seq])

            self.assertEqual(num_one, backend.add_theory_term_number(1))
            self.assertEqual(num_two, backend.add_theory_term_number(2))
            self.assertEqual(str_x, backend.add_theory_term_string("x"))
            self.assertEqual(num_one, backend.add_theory_term_symbol(parse_term("1")))
            self.assertEqual(num_two, backend.add_theory_term_symbol(parse_term("2")))
            self.assertEqual(
                fun, backend.add_theory_term_symbol(parse_term("f(1,2,x)"))
            )

            elem = backend.add_theory_element([num_one, num_two, seq, fun, lst, tup], [1, -2, 3])

            # Tests with guards and elements
            backend.add_theory_atom(fun, [], 0)
            backend.add_theory_atom_with_guard(fun, [], "=", num_one, 0)
            backend.add_theory_atom(
                backend.add_theory_term_symbol(parse_term("g(1,2)")), [], 0
            )
            backend.add_theory_atom(fseq, [elem], 0)

        atoms = list(ctl.theory_atoms)
        self.assertListEqual(
            [str(atom) for atom in atoms],
            [
                "&f(1,2,x){}",
                "&f(1,2,x){}=1",
                "&g(1,2){}",
                "&f({1,2,x}){1,2,{1,2,x},f(1,2,x),[1,2],(1,2): #aux(1),not #aux(2),#aux(3)}",
            ],
        )
        self.assertEqual(atoms[-1].elements[-1].condition, [1, -2, 3])

    def test_theory_with_guard(self):
        """
        Test observer via grounding.
        """
        ctl = Control()
        obs = TestObserverTheoryWithGuard(self)
        ctl.register_observer(obs)
        ctl.add(
            "base",
            [],
            """\
        #theory test {
            t { };
            &a/0 : t, {=}, t, head
        }.
        &a { } = a.
        """,
        )
        ctl.ground([("base", [])])
        self.assertIn("theory_term_string: a", obs.called)
        self.assertIn("theory_term_string: =", obs.called)
        self.assertIn("theory_atom_with_guard", obs.called)
        ctl.solve()



"""
Tests for solving.
"""
from unittest import TestCase
from typing import cast
from clingo import (
    Control,
    Function,
    Model,
    ModelType,
    SolveHandle,
    SolveResult,
    SymbolicAtom,
)

from .util import _MCB, _check_sat, _p


class TestSolving(TestCase):
    """
    Tests basic solving and related functions.
    """

    def setUp(self):
        self.mcb = _MCB()
        self.mit = _MCB()
        self.ctl = Control(["0"])

    def tearDown(self):
        self.mcb = None
        self.mit = None
        self.ctl = None

    def test_solve_result_str(self):
        """
        Test string representation of solve results.
        """
        ret = self.ctl.solve()
        self.assertEqual(str(ret), "SAT")
        self.assertRegex(repr(ret), "SolveResult(.*)")

    def test_model_str(self):
        """
        Test string representation of models.
        """
        self.ctl.add("base", [], "a.")
        self.ctl.ground([("base", [])])
        with cast(SolveHandle, self.ctl.solve(yield_=True)) as hnd:
            for mdl in hnd:
                self.assertEqual(str(mdl), "a")
                self.assertRegex(repr(mdl), "Model(.*)")

    def test_solve_cb(self):
        """
        Test solving using callback.
        """
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        _check_sat(
            self,
            cast(
                SolveResult,
                self.ctl.solve(on_model=self.mcb.on_model, yield_=False, async_=False),
            ),
        )
        self.assertEqual(self.mcb.models, _p(["a", "c"], ["b", "c"]))
        self.assertEqual(self.mcb.last[0], ModelType.StableModel)

    def test_solve_async(self):
        """
        Test asynchonous solving.
        """
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        with cast(
            SolveHandle,
            self.ctl.solve(on_model=self.mcb.on_model, yield_=False, async_=True),
        ) as hnd:
            _check_sat(self, hnd.get())
            self.assertEqual(self.mcb.models, _p(["a", "c"], ["b", "c"]))

    def test_solve_yield(self):
        """
        Test solving yielding models.
        """
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        with cast(
            SolveHandle,
            self.ctl.solve(on_model=self.mcb.on_model, yield_=True, async_=False),
        ) as hnd:
            for m in hnd:
                self.mit.on_model(m)
            _check_sat(self, hnd.get())
            self.assertEqual(self.mcb.models, _p(["a", "c"], ["b", "c"]))
            self.assertEqual(self.mit.models, _p(["a", "c"], ["b", "c"]))

    def test_solve_async_yield(self):
        """
        Test solving yielding models asynchronously.
        """
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        with cast(
            SolveHandle,
            self.ctl.solve(on_model=self.mcb.on_model, yield_=True, async_=True),
        ) as hnd:
            while True:
                hnd.resume()
                _ = hnd.wait()
                m = hnd.model()
                if m is None:
                    break
                self.mit.on_model(m)
            _check_sat(self, hnd.get())
            self.assertEqual(self.mcb.models, _p(["a", "c"], ["b", "c"]))
            self.assertEqual(self.mit.models, _p(["a", "c"], ["b", "c"]))

    def test_solve_interrupt(self):
        """
        Test interrupting solving.
        """
        self.ctl.add(
            "base",
            [],
            "1 { p(P,H): H=1..99 } 1 :- P=1..100.\n1 { p(P,H): P=1..100 } 1 :- H=1..99.",
        )
        self.ctl.ground([("base", [])])
        with cast(SolveHandle, self.ctl.solve(async_=True)) as hnd:
            hnd.resume()
            hnd.cancel()
            ret = hnd.get()
            self.assertTrue(ret.interrupted)

        with cast(SolveHandle, self.ctl.solve(async_=True)) as hnd:
            hnd.resume()
            self.ctl.interrupt()
            ret = hnd.get()
            self.assertTrue(ret.interrupted)

    def test_solve_core(self):
        """
        Test core retrieval.
        """
        self.ctl.add("base", [], "3 { p(1..10) } 3.")
        self.ctl.ground([("base", [])])
        ass = []
        for atom in self.ctl.symbolic_atoms.by_signature("p", 1):
            ass.append(-atom.literal)
        ret = cast(
            SolveResult, self.ctl.solve(on_core=self.mcb.on_core, assumptions=ass)
        )
        self.assertTrue(ret.unsatisfiable)
        self.assertTrue(len(self.mcb.core) > 7)

    def test_enum(self):
        """
        Test core retrieval.
        """
        self.ctl = Control(["0", "-e", "cautious"])
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        self.ctl.solve(on_model=self.mcb.on_model)
        self.assertEqual(self.mcb.last[0], ModelType.CautiousConsequences)
        self.assertEqual([self.mcb.last[1]], _p(["c"]))

        self.ctl = Control(["0", "-e", "brave"])
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        self.ctl.solve(on_model=self.mcb.on_model)
        self.assertEqual(self.mcb.last[0], ModelType.BraveConsequences)
        self.assertEqual([self.mcb.last[1]], _p(["a", "b", "c"]))

    def test_model(self):
        """
        Test functions of model.
        """

        def on_model(m: Model):
            self.assertTrue(m.contains(Function("a")))
            self.assertTrue(
                m.is_true(
                    cast(SymbolicAtom, m.context.symbolic_atoms[Function("a")]).literal
                )
            )
            self.assertFalse(m.is_true(1000))
            self.assertEqual(m.thread_id, 0)
            self.assertEqual(m.number, 1)
            self.assertFalse(m.optimality_proven)
            self.assertEqual(m.cost, [3])
            self.assertEqual(m.priority, [5])
            m.extend([Function("e")])
            self.assertSequenceEqual(m.symbols(theory=True), [Function("e")])

        self.ctl.add("base", [], "a. b. c. #minimize { 1@5,a:a; 1@5,b:b; 1@5,c:c }.")
        self.ctl.ground([("base", [])])
        self.ctl.solve(on_model=on_model)

    def test_cautious_consequences(self):
        """
        Test is_consequence function of model.
        """

        def lookup(m: Model, name: str):
            return cast(SymbolicAtom, m.context.symbolic_atoms[Function(name)]).literal

        def on_model(m: Model):
            a = lookup(m, "a")
            b = lookup(m, "b")
            c = lookup(m, "c")
            ca = m.is_consequence(a)
            cb = m.is_consequence(b)
            cc = m.is_consequence(c)
            nca = m.is_consequence(-a)
            ncb = m.is_consequence(-b)
            ncc = m.is_consequence(-c)
            self.assertTrue(ca is True)
            self.assertTrue(nca is False)
            self.assertTrue(ncb is False)
            self.assertTrue(ncc is False)
            if m.number == 1:
                self.assertTrue(ncb is None or ncb is False)
                self.assertTrue(ncc is None or ncc is False)
                self.assertTrue(cb is None or cb is False)
                self.assertTrue(cc is None or cc is False)
                self.assertTrue(cb != cc)
            if m.number == 2:
                self.assertTrue(cb is False)
                self.assertTrue(cc is False)

        self.ctl.configuration.solve.enum_mode = "cautious"
        self.ctl.add("base", [], "a. b | c.")
        self.ctl.ground([("base", [])])
        self.ctl.solve(on_model=on_model)

    def test_control_clause(self):
        """
        Test adding clauses while solving.
        """
        self.ctl.add("base", [], "1 {a; b; c} 1.")
        self.ctl.ground([("base", [])])
        with cast(
            SolveHandle,
            self.ctl.solve(on_model=self.mcb.on_model, yield_=True, async_=False),
        ) as hnd:
            for m in hnd:
                clause = []
                if m.contains(Function("a")):
                    clause.append((Function("b"), False))
                else:
                    clause.append((Function("a"), False))
                m.context.add_clause(clause)

            _check_sat(self, hnd.get())
            self.assertEqual(len(self.mcb.models), 2)

    def test_control_nogood(self):
        """
        Test adding nogoods while solving.
        """
        self.ctl.add("base", [], "1 {a; b; c} 1.")
        self.ctl.ground([("base", [])])
        with cast(
            SolveHandle,
            self.ctl.solve(on_model=self.mcb.on_model, yield_=True, async_=False),
        ) as hnd:
            for m in hnd:
                clause = []
                if m.contains(Function("a")):
                    clause.append((Function("b"), True))
                else:
                    clause.append((Function("a"), True))
                m.context.add_nogood(clause)

            _check_sat(self, hnd.get())
            self.assertEqual(len(self.mcb.models), 2)



"""
Tests control.
"""
from unittest import TestCase
from typing import cast
from clingo import Control, Function, Number, SolveResult


class TestError(Exception):
    """
    Test exception.
    """


class Context:
    """
    Simple context with some test functions.
    """

    def cb_num(self, c):
        """
        Simple test callback.
        """
        return [Number(c.number + 1), Number(c.number - 1)]

    def cb_error(self):
        """
        Simple test raising an error.
        """
        raise TestError("test")


class TestControl(TestCase):
    """
    Tests basic functions of the control object.
    """

    def test_default(self):
        """
        Test grounding with context and parameters.
        """
        ctl = Control()
        ctl.add("p.")
        ctl.ground()
        symbols = [atom.symbol for atom in ctl.symbolic_atoms]
        self.assertEqual(sorted(symbols), [Function("p", [])])

    def test_ground(self):
        """
        Test grounding with context and parameters.
        """
        ctx = Context()
        ctl = Control()
        ctl.add("part", ["c"], "p(@cb_num(c)).")
        ctl.ground([("part", [Number(1)])], ctx)
        symbols = [atom.symbol for atom in ctl.symbolic_atoms]
        self.assertEqual(
            sorted(symbols), [Function("p", [Number(0)]), Function("p", [Number(2)])]
        )

    def test_ground_error(self):
        """
        Test grounding with context and parameters.
        """
        ctx = Context()
        ctl = Control()
        ctl.add("part", ["c"], "p(@cb_error()).")
        self.assertRaisesRegex(
            TestError, "test", ctl.ground, [("part", [Number(1)])], ctx
        )

    def test_lower(self):
        """
        Test lower bounds reported during optimization.
        """
        ctl = Control(["--opt-str=usc,oll,0", "--stats=2"])
        ctl.add(
            "base",
            [],
            "1 { p(X); q(X) } 1 :- X=1..3. #minimize { 1,p,X: p(X); 1,q,X: q(X) }.",
        )
        ctl.ground([("base", [])])
        lower = []
        self.assertTrue(cast(SolveResult, ctl.solve(on_unsat=lower.append)).satisfiable)
        self.assertEqual(lower, [[1], [2], [3]])
        self.assertEqual(ctl.statistics["summary"]["lower"], [3.0])

    def test_error_handling(self):
        """
        Test basic error handling during solving.
        """
        ctl = Control()
        ctl.add("base", [], "1 {a; b} 1.")
        ctl.ground([("base", [])])

        self.assertRaises(
            ZeroDivisionError, lambda: ctl.solve(on_model=lambda m: 1 / 0)
        )

        with ctl.solve(on_model=lambda m: 1 / 0, yield_=True) as handle:
            self.assertRaises(ZeroDivisionError, lambda: [_ for _ in handle])

        # Note: currently clasp does not store and re-raise the exception in
        # asynchronous mode, so we get a runtime error instead
        with ctl.solve(on_model=lambda m: 1 / 0, async_=True) as handle:
            self.assertRaises(RuntimeError, handle.get)



"""
Tests for configuration and statistics.
"""

from unittest import TestCase
from typing import cast
from clingo import Configuration, Control


class TestConfig(TestCase):
    """
    Tests for configuration and statistics.
    """

    def test_config(self):
        """
        Test configuration.
        """
        ctl = Control(["-t", "2"])
        self.assertIn("solver", ctl.configuration.keys)
        self.assertEqual(len(ctl.configuration.solver), 2)
        self.assertIsInstance(ctl.configuration.solver[0], Configuration)
        conf = cast(Configuration, ctl.configuration.solver[0])
        self.assertIsInstance(conf.heuristic, str)
        self.assertIsInstance(conf.description("heuristic"), str)
        conf.heuristic = "berkmin"
        self.assertTrue(conf.heuristic.startswith("berkmin"))

    def test_simple_stats(self):
        """
        Test simple statistics.
        """
        ctl = Control(["-t", "2", "--stats=2"])
        ctl.add("base", [], "1 { a; b }.")
        ctl.ground([("base", [])])
        ctl.solve()
        stats = ctl.statistics
        self.assertGreaterEqual(stats["problem"]["lp"]["atoms"], 2)
        self.assertGreaterEqual(stats["solving"]["solvers"]["choices"], 1)

    def test_user_stats(self):
        """
        Test user statistics.
        """

        def on_statistics(step, accu):
            step["test"] = {"a": 0, "b": [1, 2], "c": {"d": 3}}
            accu["test"] = step["test"]
            step["test"] = {
                "a": lambda a: a + 1,
                "e": lambda a: 4 if a is None else 0,
                "b": [-1, 2, 3],
            }
            self.assertEqual(len(step["test"]), 4)
            self.assertEqual(len(step["test"]["b"]), 3)
            self.assertEqual(len(step["test"]["c"]), 1)
            self.assertIn("a", step["test"])
            self.assertEqual(sorted(step["test"]), ["a", "b", "c", "e"])
            self.assertEqual(sorted(step["test"].keys()), ["a", "b", "c", "e"])
            self.assertEqual(sorted(step["test"]["c"].items()), [("d", 3.0)])
            self.assertEqual(sorted(step["test"]["c"].values()), [3.0])

            step["test"]["b"][1] = 99
            self.assertEqual(step["test"]["b"][1], 99)
            step["test"]["b"].extend([3, 4])
            step["test"]["b"] += [3, 4]

        ctl = Control(["-t", "2", "--stats=2"])
        ctl.add("base", [], "1 { a; b }.")
        ctl.ground([("base", [])])
        ctl.solve(on_statistics=on_statistics)
        stats = ctl.statistics
        self.assertEqual(
            stats["user_step"]["test"],
            {
                "a": 1.0,
                "b": [-1.0, 99.0, 3.0, 3.0, 4.0, 3.0, 4.0],
                "c": {"d": 3.0},
                "e": 4.0,
            },
        )
        self.assertEqual(
            stats["user_accu"]["test"], {"a": 0, "b": [1, 2], "c": {"d": 3}}
        )



'''
Functions and classes to work with theory atoms.

Examples
--------
    >>> from clingo.control import Control
    >>>
    >>> ctl = Control()
    >>> ctl.add('base', [], """\\
    ... #theory example {
    ...     t { };
    ...     &a/0 : t, head
    ... }.
    ... {c}.
    ... &a { t: c }.
    ... """)
    >>> ctl.ground([('base', [])])
    >>> atm = next(ctl.theory_atoms)
    >>> print(atm)
    &a{t: c}
    >>> elm = atm.elements[0]
    >>> print(elm)
    t: c
'''

from typing import List, Optional, Tuple
from functools import total_ordering

from ._internal import _c_call, _c_call2, _lib, _str, _to_str
from .core import OrderedEnum

__all__ = ["TheoryAtom", "TheoryElement", "TheoryTerm", "TheoryTermType"]


class TheoryTermType(OrderedEnum):
    """
    Enumeration of theory term types.
    """

    Function = _lib.clingo_theory_term_type_function
    """
    For a function theory terms.
    """
    List = _lib.clingo_theory_term_type_list
    """
    For list theory terms.
    """
    Number = _lib.clingo_theory_term_type_number
    """
    For numeric theory terms.
    """
    Set = _lib.clingo_theory_term_type_set
    """
    For set theory terms.
    """
    Symbol = _lib.clingo_theory_term_type_symbol
    """
    For symbolic theory terms (symbol here means the term is a string).
    """
    Tuple = _lib.clingo_theory_term_type_tuple
    """
    For tuple theory terms.
    """


@total_ordering
class TheoryTerm:
    """
    `TheoryTerm` objects represent theory terms.

    Theory terms have a readable string representation, implement Python's rich
    comparison operators, and can be used as dictionary keys.
    """

    def __init__(self, rep, idx):
        self._rep = rep
        self._idx = idx

    def __hash__(self):
        return self._idx

    def __eq__(self, other):
        return self._idx == other._idx

    def __lt__(self, other):
        return self._idx < other._idx

    def __str__(self):
        return _str(
            _lib.clingo_theory_atoms_term_to_string_size,
            _lib.clingo_theory_atoms_term_to_string,
            self._rep,
            self._idx,
        )

    def __repr__(self):
        return f"TheoryTerm({self._rep!r})"

    @property
    def arguments(self) -> List["TheoryTerm"]:
        """
        The arguments of the term (for functions, tuples, list, and sets).
        """
        args, size = _c_call2(
            "clingo_id_t*",
            "size_t",
            _lib.clingo_theory_atoms_term_arguments,
            self._rep,
            self._idx,
        )
        return [TheoryTerm(self._rep, args[i]) for i in range(size)]

    @property
    def name(self) -> str:
        """
        The name of the term (for symbols and functions).
        """
        return _to_str(
            _c_call("char*", _lib.clingo_theory_atoms_term_name, self._rep, self._idx)
        )

    @property
    def number(self) -> int:
        """
        The numeric representation of the term (for numbers).
        """
        return _c_call(
            "int", _lib.clingo_theory_atoms_term_number, self._rep, self._idx
        )

    @property
    def type(self) -> TheoryTermType:
        """
        The type of the theory term.
        """
        type_ = _c_call(
            "clingo_theory_term_type_t",
            _lib.clingo_theory_atoms_term_type,
            self._rep,
            self._idx,
        )
        return TheoryTermType(type_)


@total_ordering
class TheoryElement:
    """
    Class to represent theory elements.

    Theory elements have a readable string representation, implement Python's rich
    comparison operators, and can be used as dictionary keys.
    """

    def __init__(self, rep, idx):
        self._rep = rep
        self._idx = idx

    def __hash__(self):
        return self._idx

    def __eq__(self, other):
        return self._idx == other._idx

    def __lt__(self, other):
        return self._idx < other._idx

    def __str__(self):
        return _str(
            _lib.clingo_theory_atoms_element_to_string_size,
            _lib.clingo_theory_atoms_element_to_string,
            self._rep,
            self._idx,
        )

    def __repr__(self):
        return f"TheoryElement({self._rep!r})"

    @property
    def condition(self) -> List[int]:
        """
        The condition of the element in form of a list of program literals.
        """
        cond, size = _c_call2(
            "clingo_literal_t*",
            "size_t",
            _lib.clingo_theory_atoms_element_condition,
            self._rep,
            self._idx,
        )
        return [cond[i] for i in range(size)]

    @property
    def condition_id(self) -> int:
        """
        Each condition has an id, which is a temporary program literal. This id
        can be passed to `clingo.propagator.PropagateInit.solver_literal` to
        obtain a corresponding solver literal.
        """
        return _c_call(
            "clingo_literal_t",
            _lib.clingo_theory_atoms_element_condition_id,
            self._rep,
            self._idx,
        )

    @property
    def terms(self) -> List[TheoryTerm]:
        """
        The tuple of the element.
        """
        terms, size = _c_call2(
            "clingo_id_t*",
            "size_t",
            _lib.clingo_theory_atoms_element_tuple,
            self._rep,
            self._idx,
        )
        return [TheoryTerm(self._rep, terms[i]) for i in range(size)]


@total_ordering
class TheoryAtom:
    """
    Class to represent theory atoms.

    Theory atoms have a readable string representation, implement Python's rich
    comparison operators, and can be used as dictionary keys.
    """

    def __init__(self, rep, idx):
        self._rep = rep
        self._idx = idx

    def __hash__(self):
        return self._idx

    def __eq__(self, other):
        return self._idx == other._idx

    def __lt__(self, other):
        return self._idx < other._idx

    def __str__(self):
        return _str(
            _lib.clingo_theory_atoms_atom_to_string_size,
            _lib.clingo_theory_atoms_atom_to_string,
            self._rep,
            self._idx,
        )

    def __repr__(self):
        return f"TheoryAtom({self._rep!r})"

    @property
    def elements(self) -> List[TheoryElement]:
        """
        The elements of the atom.
        """
        elems, size = _c_call2(
            "clingo_id_t*",
            "size_t",
            _lib.clingo_theory_atoms_atom_elements,
            self._rep,
            self._idx,
        )
        return [TheoryElement(self._rep, elems[i]) for i in range(size)]

    @property
    def guard(self) -> Optional[Tuple[str, TheoryTerm]]:
        """
        The guard of the atom or None if the atom has no guard.
        """
        if not _c_call(
            "bool", _lib.clingo_theory_atoms_atom_has_guard, self._rep, self._idx
        ):
            return None

        conn, term = _c_call2(
            "char*",
            "clingo_id_t",
            _lib.clingo_theory_atoms_atom_guard,
            self._rep,
            self._idx,
        )

        return (_to_str(conn), TheoryTerm(self._rep, term))

    @property
    def literal(self) -> int:
        """
        The program literal associated with the atom.
        """
        return _c_call(
            "clingo_literal_t",
            _lib.clingo_theory_atoms_atom_literal,
            self._rep,
            self._idx,
        )

    @property
    def term(self) -> TheoryTerm:
        """
        The term of the atom.
        """
        term = _c_call(
            "clingo_id_t", _lib.clingo_theory_atoms_atom_term, self._rep, self._idx
        )
        return TheoryTerm(self._rep, term)
